/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg;

import java.io.File;
import java.util.List;
import java.util.concurrent.ExecutionException;

import com.aragost.javahg.RepositoryConfiguration.CachePolicy;
import com.aragost.javahg.internals.ExtensionManager;
import com.aragost.javahg.internals.Server;
import com.aragost.javahg.internals.Utils;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.CacheStats;
import com.google.common.cache.LoadingCache;

/**
 * This is a standard Repository with no bundle overlaid
 */
public class BaseRepository extends Repository {

    private RepositoryConfiguration configuration;

    private final File directory;

    private Server server;
    private LoadingCache<String, Changeset> nodeChangeSetCache;

    BaseRepository(RepositoryConfiguration conf, File directory, boolean performInit, String cloneUrl) {
        List<String> extensionArgs = ExtensionManager.getInstance().process(conf.getExtensionClasses());
        this.nodeChangeSetCache = createCache(conf.getCachePolicy());
        this.configuration = conf;
        this.directory = Utils.resolveSymlinks(directory);

        String hgBin = this.configuration.getHgBin();
        File hgBinFile = new File(hgBin);
        if (hgBinFile.isAbsolute() && !hgBinFile.exists()) {
            throw new IllegalArgumentException("" + hgBin + " does not exist");
        }

        this.server = new Server(hgBin);
        this.server.setErrorAction(this.configuration.getCodingErrorAction());
        this.server.incrementRefCount();
        if (performInit) {
            this.server.initMecurialRepository(directory);
        } else if (cloneUrl != null) {
            this.server.execHgCommand(null, "clone", cloneUrl, directory.getAbsolutePath());
        }
        this.server.start(directory, extensionArgs);
    }

    @Override
    public Changeset changeset(String node) {
        if (node.equals(Changeset.NULL_ID)) {
            return null;
        } else {
            try {
                return this.nodeChangeSetCache.get(node);
            } catch (ExecutionException e) {
                throw Utils.asRuntime(e);
            }
        }
    }

    @Override
    public Server getServer() {
        return this.server;
    }

    @Override
    public File getDirectory() {
        return this.directory;
    }

    RepositoryConfiguration getConfiguration() {
        return configuration;
    }

    @Override
    public BaseRepository getBaseRepository() {
        return this;
    }

    private LoadingCache<String, Changeset> createCache(CachePolicy cachePolicy) {
        CacheLoader<String, Changeset> cacheLoader = new CacheLoader<String, Changeset>() {
            @Override
            public Changeset load(String key) throws Exception {
                return new Changeset(BaseRepository.this, key);
            }
        };
        CacheBuilder<Object, Object> cacheBuilder = CacheBuilder.newBuilder();
        switch (cachePolicy) {
        case STRONG:
            break;
        case SOFT:
            cacheBuilder.softValues();
            break;
        case WEAK:
            cacheBuilder.weakValues();
            break;
        case NONE:
            // Consider not to have any cache at all
            cacheBuilder.maximumSize(0);
            break;
        }
        return cacheBuilder.build(cacheLoader);
    }

    public CacheStats getCacheStats() {
        return this.nodeChangeSetCache.stats();
    }

}
