/*
 * #%L
 * JavaHg
 * %%
 * Copyright (C) 2011 aragost Trifork ag
 * %%
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 * #L%
 */
package com.aragost.javahg.commands;

import java.io.IOException;
import java.util.Collections;
import java.util.List;

import com.aragost.javahg.Bookmark;
import com.aragost.javahg.Repository;
import com.aragost.javahg.commands.flags.BookmarksCommandFlags;
import com.aragost.javahg.internals.HgInputStream;
import com.aragost.javahg.internals.RuntimeIOException;
import com.google.common.collect.Lists;

/**
 * Command class for executing <tt>hg bookmarks</tt>.
 */
public class BookmarksCommand extends BookmarksCommandFlags {

    private static final byte[] NO_BOOKMARKS = "no bookmarks set\n".getBytes();

    public BookmarksCommand(Repository repository) {
        super(repository);
        withDebugFlag();
    }

    /**
     * 
     * @return all bookmarks
     */
    public List<Bookmark> list() {
        Repository repo = getRepository();
        HgInputStream stream = launchStream();
        try {
            if (stream.match(NO_BOOKMARKS)) {
                return Collections.emptyList();
            }
            List<Bookmark> result = Lists.newArrayList();
            while (stream.peek() != -1) {
                stream.mustMatch(' ');
                boolean active = stream.read() == '*';
                stream.mustMatch(' ');
                String name = stream.textUpTo(' ');
                stream.upTo(':');
                String node = stream.textUpTo('\n');
                Bookmark bookmark = new Bookmark(repo.changeset(node), name, active);
                result.add(bookmark);
            }
            return result;
        } catch (IOException e) {
            throw new RuntimeIOException(e);
        }
    }

    /**
     * Create a new bookmark
     * 
     * @param name
     */
    public void create(String name) {
        launchString(name);
    }

    /**
     * Delete a bookmark
     * 
     * @param name
     */
    public void delete(String name) {
        launchString("--delete", name);
    }

    /**
     * Rename a bookmark
     * 
     * @param oldName
     * @param newName
     */
    public void rename(String oldName, String newName) {
        launchString("--rename", oldName, newName);
    }
}
